@@ -1,7 +1,7 @@
 from typing import Union, Dict, Any, List, Tuple
 import pandas as pd
 from tqdm.auto import tqdm
-from .utils import divvy_workload, get_num_workers, multiprocess
+from .utils import divvy_workload, get_num_workers, process_queue
 from pydantic import BaseModel
 from multiprocessing import Pool
 from copy import deepcopy
@@ -53,7 +53,7 @@ def parse_string_with_manual_tokens(input_string: str, tokens_dictionary: Dict[s
 
     # Loop over rows
     for raw_row_text in (p := tqdm(raw_row_texts_list, disable=disable_progress_bar)):
-        p.set_description('Reading and parsing (step 1 of 2)')
+        p.set_description('Reading and parsing')
         row_buffer: Dict[str, Any] = dict()
 
         # Look for each key in this row
@@ -133,7 +133,7 @@ def parse_string_with_key_value_delimiters(input_string: str, left_token: str =
     else:
         # Serial processing
         for chunk in (p1 := tqdm(raw_row_texts_list, disable=disable_progress_bar)):
-            p1.set_description('Scanning file (step 1 of 2)')
+            p1.set_description('Scanning file')
             row_buffers_list.append(key_value_extraction_lambda(chunk, left_token=left_token, right_token=right_token,
                                                                 key_value_delimiter=key_value_delimiter))
 
@@ -148,13 +148,14 @@ def parse_file_with_key_value_delimiters(path: str, *args, **kwargs):
 
 
 def parse_string_with_embedded_json_unprocessed_dicts(input_string: str, embedded_json_line_prefix: str = None,
-                                                      limit: int = None,
+                                                      limit: int = None, ignore_decode_errors: bool = False,
                                                       end_of_line: str = '\n') -> List[Dict[Any, Any]]:
     """
     Extracts embedded json from log files
 
     :param input_string: string (e.g. logs file) to be parsed
     :param embedded_json_line_prefix: marker string at the start of every embedded json line
+    :param ignore_decode_errors: if True, silently ignores rows that don't parse (e.g. if pulling from incomplete row)
     :param end_of_line: substring between rows
     :param limit: maximum number of rows to process
     """
@@ -173,7 +174,14 @@ def parse_string_with_embedded_json_unprocessed_dicts(input_string: str, embedde
         embedded_with_trailing_data = embedded_with_trailing_data[:limit]
     embedded_clean = [x.split(end_of_line)[0] if end_of_line in x else x for x in embedded_with_trailing_data]
 
-    return [json.loads(s) for s in embedded_clean]
+    return_list: List[Dict[Any, Any]] = []
+    for s in embedded_clean:
+        try:
+            return_list.append(json.loads(s))
+        except json.JSONDecodeError as e:
+            if not ignore_decode_errors:
+                raise e
+    return return_list
 
 
 def parse_string_with_embedded_json(input_string: str, embedded_json_line_prefix: str = None,
@@ -269,7 +277,7 @@ def list_of_dict_to_dataframe(data_point_dicts, parallelize_processing: Union[bo
                 print(f'Reshaping data (step 1 of 2): processing in parallel with {num_reshape_workers} workers')
             batches_of_dictionaries: List[List[Any]] = divvy_workload(num_workers=num_reshape_workers,
                                                                       tasks=data_point_dicts)
-            list_of_dataframes: List[pd.DataFrame] = multiprocess(
+            list_of_dataframes: List[pd.DataFrame] = process_queue(
                 data_frame_init_wrapper,
                 batches_of_dictionaries,
                 pool_function='map',
@@ -277,18 +285,18 @@ def list_of_dict_to_dataframe(data_point_dicts, parallelize_processing: Union[bo
                 num_workers=num_reshape_workers,
             )
             if not disable_progress_bar:
-                print('Reshaping data (step 1 of 2): attaching subframes')
+                print('Reshaping data (step 2 of 2): attaching subframes')
             return pd.concat(list_of_dataframes, ignore_index=True)
 
     if not suppress_deprecation_warning:
         print(f"You are using a very slow method for reshaping the data")
         print(f"Recommendation: use list_of_dict_to_dataframe(..., pandas_automatic=True)")
-        print(f"To suppress thing warning, pass: use list_of_dict_to_dataframe(..., suppress_deprecation_warning=True)")
+        print(f"To suppress this warning, pass: use list_of_dict_to_dataframe(..., suppress_deprecation_warning=True)")
 
     if parallelize_processing and (num_reshape_workers > 1):
         batches: List[List[Any]] = divvy_workload(num_workers=num_reshape_workers, tasks=data_point_dicts)
         if not disable_progress_bar:
-            print(f'Reshaping data (step 1 of 2): processing in parallel with {num_reshape_workers} workers')
+            print(f'Reshaping data: processing in parallel with {num_reshape_workers} workers')
         with Pool(num_reshape_workers) as pool:
             dataframes: List[pd.DataFrame] = pool.map(func=dicts_to_dataframe, iterable=batches)
         df: pd.DataFrame = pd.concat(dataframes, ignore_index=True)
@@ -296,7 +304,7 @@ def list_of_dict_to_dataframe(data_point_dicts, parallelize_processing: Union[bo
         df: pd.DataFrame = pd.DataFrame()
         # THIS LOOP IS SO SLOW - you should use pandas_automatic=True
         for chunk_buffer in (p2 := tqdm(data_point_dicts, disable=disable_progress_bar)):
-            p2.set_description('Reshaping data (step 2 of 2):')
+            p2.set_description('Reshaping data:')
             if chunk_buffer:
                 df: pd.DataFrame = pd.concat([df, pd.DataFrame(chunk_buffer, index=[-1])], ignore_index=True)
     return df
