@@ -4,15 +4,16 @@ import matplotlib.pyplot as plt
 import numpy as np
 import pandas as pd
 from .config import Style
-from .utils import PickleUtils, Rosetta, make_dict, get_num_workers
+from .utils import PickleUtils, Rosetta, make_dict, get_num_workers, process_queue
 from .data_quality import basis_quality_checks, basis_quality_plots, fill_ratio
 from copy import deepcopy
 import statsmodels.api as sm
-from .plotting import visualize_x_y, visualize_1d_distribution, visualize_surface, visualize_embedded_surface
+from .plotting import visualize_x_y, visualize_1d_distribution, visualize_surface, \
+    visualize_embedded_surface, visualize_hist2d, visualize_x_y_input_interpreter
 import pathlib
 from pydantic import BaseModel
 from sklearn.feature_selection import SelectKBest, chi2
-from multiprocessing import Pool
+from tqdm.auto import tqdm
 
 
 class SVD(BaseModel):
@@ -34,6 +35,12 @@ class VectorMultiset(PickleUtils, Style, Rosetta):
         """ The length of the vector set is the length of the data frame """
         return len(self.data)
 
+    def __add__(self, other):
+        """ Adding items derived from MultiSets concatenates their dataframes (keeping metadat from left side) """
+        to_return = deepcopy(self)
+        to_return.data = pd.concat([self.data, other.data])
+        return to_return
+
     def values(self, feature: str, cumulative: bool = False, *args) -> List[Any]:
         """ Retrieves a particular data feature by attribute name. Additional args unpack deeper
 
@@ -41,6 +48,8 @@ class VectorMultiset(PickleUtils, Style, Rosetta):
         :param cumulative: whether to apply a cumulative sum (default = False)
         :return: Extracted data
         """
+        if feature not in self.data.keys():
+            raise ValueError(f"Feature {feature} not in known keys: {self.data.keys().tolist()}")
         values: pd.Series = self.data.loc[:, feature]
         if args:
             for unpack_next in args:
@@ -148,11 +157,22 @@ class VectorMultiset(PickleUtils, Style, Rosetta):
         style: Style = kwargs.pop('style', Style(**self.dict()))
         return visualize_x_y(x_data, y_data, style=style, **kwargs)
 
+    def viz2d(self, *args, **kwargs) -> plt.Figure:
+        """
+        For convenience, allow direct access to visualize_x_y_input_interpreter
+
+        :param args: visualize_x_y_input_interpreter positional inputs
+        :param kwargs: visualize_x_y_input_interpreter keyword argument inputs
+        :return: figure handle
+        """
+        kwargs.setdefault('style', Style(**self.dict()))
+        return visualize_x_y_input_interpreter(*args, **kwargs)
+
     def scatter(self, *args, **kwargs) -> plt.Figure:
         """ Creates a 2D scatter plot of x and y data (wraps visualize_x_y) """
         if isinstance(self, VectorSequence) and (len(args) == 1):
             args: tuple = (self.basis_col_name, args[0])  # sequences we can infer the basis for the x-axis
-        return self.visualize_x_y(types='line', *args, **kwargs)
+        return self.visualize_x_y(types='scatter', *args, **kwargs)
 
     def plot(self, *args, **kwargs) -> plt.Figure:
         """ Creates a 2D line plot of x and y data (wraps visualize_x_y) """
@@ -173,6 +193,25 @@ class VectorMultiset(PickleUtils, Style, Rosetta):
         kwargs.setdefault('title', self.translate(self.name_root))
         return visualize_1d_distribution(data, style=Style(**self.dict()), **kwargs)
 
+    def hist2d(self, *args, **kwargs) -> plt.Figure:
+        """ Creates a 2D histogram plot of x and y data (wraps visualize_hist2d) """
+        if isinstance(self, VectorSequence) and (len(args) == 1):
+            args: tuple = (self.basis_col_name, args[0])  # sequences we can infer the basis for the x-axis
+
+        # If the first argument is a string, replace it with column data
+        if isinstance(args[0], str):
+            kwargs.setdefault('xlabel', self.translate(args[0]))
+            args = self.data.loc[:, args[0]].tolist(), *args[1:]
+
+        # If the second argument is a string, replace it with column data
+        if isinstance(args[1], str):
+            kwargs.setdefault('ylabel', self.translate(args[1]))
+            extra_args = args[2:] if len(args) > 0 else tuple()
+            args = args[0], self.data.loc[:, args[1]].tolist(), *extra_args
+
+        kwargs.setdefault('title', self.name_root)
+        return visualize_hist2d(*args, **kwargs)
+
     def surface(self, x_name: str, y_name: str, z_name: str, **kwargs) -> plt.Figure:
         """ Plot a surface from the data. Useful kwargs: [cumulative, density, bins]
 
@@ -270,6 +309,53 @@ class VectorMultiset(PickleUtils, Style, Rosetta):
         target_feature_data = self.data.loc[:, target_feature_name].to_numpy()
         return SelectKBest(chi2, k=k_best).fit_transform(input_feature_data, target_feature_data, **kwargs)
 
+    def cast_to_numeric(self, columns: Union[List[str], str] = None, errors: str = "ignore",
+                        inplace: bool = True) -> Union[None, Any]:
+        """
+        Helper function that converts the data to numeric types
+
+        :param columns: a string or list of strings of column names. If not provided, will try on all columns
+        :param errors: passed to `pandas.to_numeric()`, options are: {‘ignore’, ‘raise’, ‘coerce’}
+        :param inplace: whether to transform the data inplace (default) or return a fresh object
+        """
+        if not columns:
+            columns = list(self.data.keys())
+        if isinstance(columns, str):
+            columns = [columns]
+
+        df: pd.DataFrame = deepcopy(self.data)
+        for col in columns:
+            df[col] = pd.to_numeric(df.loc[:, col], errors=errors)
+
+        if inplace:
+            self.data = df
+        else:
+            return self.__class__(data=df, **{k: v for k, v in self.dict().items() if k != 'data'})
+
+    def __init__(self, **kwargs):
+        super().__init__(**kwargs)
+        if not kwargs.get('disable_auto_conversion_to_numeric', False) and (self.data is not None):
+            self.cast_to_numeric()
+
+    def drop_col_types(self, drop_types: Union[type, List[type]], inplace: bool = True) -> Union[Any, None]:
+        """
+        Helper function to drop columns of a particular type(s)
+
+        :param drop_types: the types to be dropped
+        :param inplace: whether the columns should be dropped inplace or if a new instance should be returned
+        :return: None if inplace=True, otherwise a copy of the object with dropped columns
+        """
+        if not isinstance(drop_types, (list, tuple)):
+            drop_types = [drop_types]
+        target_cols: List[str] = [k for k in self.data.keys() if isinstance(self.data.loc[:, k][0], drop_types)]
+        if inplace:
+            self.data.drop(columns=target_cols, inplace=True)
+            return None
+        else:
+            to_return: Any = deepcopy(self)
+            to_return.drop_col_types(drop_types=drop_types, inplace=True)
+            return to_return
+
 
 class SlidingWindowResults(VectorMultiset):
     """ Results from Sequence.sliding_window(), which is a VectorMultiset with extra context baked in """
@@ -374,7 +460,7 @@ class VectorSequence(VectorMultiset):
         return self.data.loc[:, self.basis_col_name].tolist()
 
     def basis_quality_checks(self) -> Tuple[bool, str]:
-        """ Checks basis data quality and returns both a self.data.loc[:, self.basis_col_name]True/False flag and a string with an explanation
+        """ Checks basis data quality and returns both a True/False flag and a string with an explanation
 
         :return: True if OK, and a string with explanation either way
         """
@@ -388,12 +474,12 @@ class VectorSequence(VectorMultiset):
         result, _ = self.basis_quality_checks()
         return result
 
-    def basis_quality_plots(self) -> List[plt.Figure]:
+    def basis_quality_plots(self, **kwargs) -> List[plt.Figure]:
         """ Creates a series of plots showing the basis data quality (missing data, nans, timeseries gaps, etc)
 
         :return: figure handles for several plots
         """
-        return basis_quality_plots(self.basis())
+        return basis_quality_plots(self.basis(), **kwargs)
 
     def fill_ratio(self) -> float:
         """ Checks for how many data points are observed for the expected number of data points in a uniform basis.
@@ -406,9 +492,11 @@ class VectorSequence(VectorMultiset):
         return fill_ratio(self.data.loc[:, self.basis_col_name])
 
     def slice(self, start_at: Any = None, stop_at: Any = None, inplace: bool = False, reset_index: bool = True,
-              return_type: str = 'VectorSequence') -> Union[None, pd.DataFrame, Tuple[Iterable, Iterable], Any]:
+              return_type: str = None) -> Union[None, pd.DataFrame, Tuple[Iterable, Iterable], Any]:
         """ Slices the VectorSequence according to the basis
 
+        # TODO have slice return self type for convenience
+
         :param start_at: Start point (can be None)
         :param stop_at: Stop point (can be None)
         :param inplace: slice inplace or return the result (default --> return)
@@ -427,11 +515,16 @@ class VectorSequence(VectorMultiset):
             in_range.reset_index(drop=True, inplace=True)
         if inplace:
             self.data = in_range
+
+        if not return_type:
+            return self.__class__(basis_col_name=self.basis_col_name, name_root=self.name_root, data=in_range)
         else:
             if 'sequence' in (return_type_lower := return_type.lower()):
                 return VectorSequence(basis_col_name=self.basis_col_name, name_root=self.name_root, data=in_range)
             elif 'dataframe' in return_type_lower:
                 return in_range
+            elif 'timeseries' in return_type_lower:
+                return Timeseries(basis_col_name=self.basis_col_name, name_root=self.name_root, data=in_range)
             else:
                 raise ValueError(f"Unknown return type: {return_type}")
 
@@ -511,16 +604,21 @@ class VectorSequence(VectorMultiset):
     #                                       vv sequence vv       vv args vv          vv kwargs vv
     def sliding_window(self, function: Callable[[Any, Union[Tuple[Any], List[Any]], Dict[str, Any]], Dict[str, Any]],
                        window_widths: List[float] = None, window_starts: List[Any] = None, step_size: float = None,
-                       parallelize_sliding_window: Union[bool, int] = True, *args, **kwargs) -> SlidingWindowResults:
+                       parallelize_sliding_window: Union[bool, int] = True, allow_shorter_windows: bool = False,
+                       disable_sliding_window_progress_bar=None,
+                       limit: int = None, *args, **kwargs) -> SlidingWindowResults:
         """ Apply function in a sliding window over the sequence
 
         :param function: callable to be applied
-        :param window_widths: list of window widths to use
+        :param window_widths: list of window widths to use (or a single value if only one width desired)
         :param window_starts: list of starting points for the windows
         :param step_size: how far apart to space windows
+        :param allow_shorter_windows: whether to allow windows that are not the full length (disabled by default)
         :param args: positional arguments for the function
         :param kwargs: keyword arguments for the function
-        :param parallelize_sliding_window: Whether to use multiprocessing for the sliding window
+        :param parallelize_sliding_window: whether to use multiprocessing for the sliding window
+        :param disable_sliding_window_progress_bar: pass True to disable progress bar (only applies to serial mode)
+        :param limit: set a limit for how many windows to evaluate, takes the first N
         :return: SlidingWindowResults (see the dataframe attribute for results)
         """
 
@@ -532,6 +630,9 @@ class VectorSequence(VectorMultiset):
 
         basis: Tuple[float] = (self.data.loc[:, self.basis_col_name].tolist())
 
+        if isinstance(window_widths, (float, int)):
+            window_widths = [window_widths]
+
         if not window_widths:
             duration: float = max(self.data.loc[:, self.basis_col_name]) - min(self.data.loc[:, self.basis_col_name])
             window_widths: List[float] = [duration / x for x in range(20, 401, 20)]  # TODO: move vals to vars
@@ -541,27 +642,36 @@ class VectorSequence(VectorMultiset):
         for i, window_width in enumerate(window_widths):
             if not window_starts:
                 if step_size:
-                    window_starts: List[float] = list(np.arange(min(basis), max(basis) - window_width, step_size))
+                    if allow_shorter_windows:
+                        window_starts: List[float] = list(np.arange(min(basis), max(basis), step_size))
+                    else:
+                        window_starts: List[float] = list(np.arange(min(basis), max(basis) - window_width, step_size))
                 else:
                     window_starts: List[float] = list(np.arange(min(basis), max(basis) - window_width, window_width))
             list_of_start_and_width_tuples += [(start_time, window_width) for start_time in window_starts]
 
-        # If parallelize_sliding_window != False, run in parallel using starmap() from multiprocessing library `Pool`
-        if parallelize_sliding_window:
-            num_workers: int = get_num_workers(parallelize_sliding_window)
+        if not list_of_start_and_width_tuples:
+            raise ValueError("No windows for sliding analysis - check your sizes and start times")
 
-            # Create a Pool and process the evaluations in parallel
-            with Pool(num_workers) as pool:
-                evaluations: List[Dict[Any, Any]] = pool.starmap(
-                    func=self.evaluate_over_window,
-                    iterable=[(function, start, width, args, kwargs) for start, width in list_of_start_and_width_tuples]
-                )
+        if limit and limit < len(list_of_start_and_width_tuples):
+            list_of_start_and_width_tuples = list_of_start_and_width_tuples[:limit]
+
+        # If parallelize_sliding_window != False, run in parallel using starmap() from multiprocessing library `Pool`
+        num_workers: int = get_num_workers(parallelize_sliding_window)
+        if parallelize_sliding_window and (num_workers > 1):
+            evaluations: List[Dict[Any, Any]] = process_queue(
+                func=self.evaluate_over_window,
+                iterable=[(function, start, width, args, kwargs) for start, width in list_of_start_and_width_tuples],
+                pool_function='starmap',
+                batching=False,
+                num_workers=num_workers
+            )
 
         # If parallelize_sliding_window == False (or None) use `for` loop in list comprehension -> serial processing
         else:
             evaluations: List[Dict[Any, Any]] = [self.evaluate_over_window(
                 function=function, start_at=start, window_width=width, args=args, kwargs=kwargs
-            ) for start, width in list_of_start_and_width_tuples]
+            ) for start, width in tqdm(list_of_start_and_width_tuples, disable=disable_sliding_window_progress_bar)]
 
         return SlidingWindowResults(window_width_col_name='window_width', window_start_col_name='window_start',
                                     data=pd.DataFrame(evaluations), name_root=self.name_root)
@@ -731,6 +841,11 @@ class VectorSequence(VectorMultiset):
         return correlation_matrix(self.data, **kwargs)
 
 
+class Timeseries(VectorSequence):
+    """ Thin wrapper for VectorSequence in the context of time  """
+    basis_col_name: str = 'timestamp'
+
+
 ################
 # Helpers
 ################
